# Air at sea level conditions h=0 (m)
const ρ0 = 1.225  # Density at sea level (kg/m3)
const p0 = 101325.0  # Pressure at sea level (Pa)
const a0 = 340.293990543  # Sound speed at sea level (m/s)


"""
    qc2cas(qc)

Calculate calibrated airspeed from ASI (Air Speed indicator), differential
pressure between impact pressure and static pressure.

``qc = p_t - p_s``

# References

1. Ward, D. T. (1993). Introduction to flight test engineering. Elsevier Science Ltd.
   (page 13, formula 2.13)
"""
function qc2cas(qc)
    cas = sqrt((2*γ_AIR*p0) / ((γ_AIR - 1) * ρ0) * ((qc / p0 + 1) ^ ((γ_AIR - 1) / γ_AIR) - 1))
    return cas
end


"""
    qc2tas(qc, ρ, p)

Calculate true airspeed from ASI (Air Speed indicator), differential
pressure between impact pressure and static pressure (qc = p_t - p_s), rho
and p.

# References

1. Ward, D. T. (1993). Introduction to flight test engineering. Elsevier Science Ltd.
   (page 12, based on formula 2.11)
"""
function qc2tas(qc, ρ, p)
    tas2 = (2*γ_AIR*p) / ((γ_AIR - 1) * ρ0) * ((qc/p + 1) ^ ((γ_AIR - 1) / γ_AIR) - 1) * ρ0/ρ
    tas = sqrt(tas2)
    return tas
end


"""
    qc2eas(qc, p)

Calculate equivalent airspeed from ASI (Air Speed indicator), differential
pressure between impact pressure and static pressure (qc = p_t - p_s) and p.
# References

1. Ward, D. T. (1993). Introduction to flight test engineering. Elsevier Science Ltd.
"""
function qc2eas(qc, p)
    eas = sqrt((2*γ_AIR*p) / ((γ_AIR - 1) * ρ0) * ((qc/p + 1) ^ ((γ_AIR - 1) / γ_AIR) - 1))
    return eas
end


"""
    tas2eas(tas, ρ)

Calculate equivalent airspeed from true airspeed and density at current
altitude (ρ).

# References

1. Ward, D. T. (1993). Introduction to flight test engineering. Elsevier Science Ltd.
   (page 13, formula 2.15)
"""
function tas2eas(tas, ρ)
    eas = tas * sqrt(ρ / ρ0)
    return eas
end


"""
    eas2tas(qc, ρ)

Calculate true airspeed from equivalent airspeed and density at current
altitude (ρ).

# References

1. Ward, D. T. (1993). Introduction to flight test engineering. Elsevier Science Ltd.
   (page 13, formula 2.15)
"""
function eas2tas(eas, ρ)
    tas = eas / sqrt(ρ / ρ0)
    return tas
end


"""
    cas2eas(cas, ρ, p)

Calculate equivalent airspeed from calibrated airspeed, density (ρ) and
pressure (p) at the current altitude.
"""
function cas2eas(cas, ρ, p)
    tas = cas2tas(cas, ρ, p)
    eas = tas2eas(tas, ρ)
    return eas
end


"""
    eas2cas(eas, ρ, p)

Calculate calibrated airspeed from equivalent airspeed, density (ρ) and
pressure (p) at the current altitude.
"""
function eas2cas(eas, ρ, p)
    tas = eas2tas(eas, ρ)
    cas = tas2cas(tas, ρ, p)
    return cas
end


"""
    cas2tas(cas, ρ, p)

Calculate true airspeed from calibrated airspeed, density (ρ) and pressure (p)
at the current altitude.
"""
function cas2tas(cas, ρ, p)

    a = sqrt(γ_AIR * p / ρ)

    temp = (cas*cas * (γ_AIR - 1.0)/(2.0 * a0*a0) + 1.0) ^ (γ_AIR / (γ_AIR - 1.0))
    temp = (temp - 1.0) * (p0 / p)
    temp = (temp + 1.0) ^ ((γ_AIR - 1.0) / γ_AIR) - 1.0

    tas = sqrt(2.0 * a*a / (γ_AIR - 1.0) * temp)

    return tas
end


"""
    tas2cas(cas, ρ, p)

Calculate true airspeed from calibrated airspeed, density (ρ) and pressure (p)
at the current altitude.
"""
function tas2cas(tas, ρ, p)

    a = sqrt(γ_AIR * p / ρ)

    temp = (tas*tas * (γ_AIR - 1.0)/(2.0 * a*a) + 1.0) ^ (γ_AIR / (γ_AIR - 1.0))
    temp = (temp - 1.0) * (p / p0)
    temp = (temp + 1.0) ^ ((γ_AIR - 1.0) / γ_AIR) - 1.0

    cas = sqrt(2.0 * a0*a0 / (γ_AIR - 1) * temp)

    return cas
end


"""
    incompressible_qinf(tas, ρ)

Calculate incompressible dynamic pressure from true airspeed (tas) and density
(ρ) at current altitude.

# References

1. Ward, D. T. (1993). Introduction to flight test engineering. Elsevier Science Ltd.
   (page 13, formula 2.14)
"""
function incompressible_qinf(tas, ρ)
    return 0.5 * ρ * tas*tas
end


"""
    compressible_qinf(tas, p, a)

Calculate compressible dynamic pressure from Mach number and static
pressure (p)

Two different models are used depending on the Mach number:

- Subsonic case: Bernouilli's equation compressible form.
- Supersonic case: to be implemented.

# References

1. Ward, D. T. (1993). Introduction to flight test engineering. Elsevier Science Ltd.
   (page 12)
2. Fundamentals of Aerdynamics, 5th edition, J.D.Anderson Jr (page 550)
"""
function compressible_qinf(M, p)

    if M < 1
        pt = p * (1 + (γ_AIR - 1) / 2 * M*M) ^ (γ_AIR / (γ_AIR - 1))
    else    # (M >= 1)
        pt = p * ((1 - γ_AIR + (2 * γ_AIR * M^2))/(γ_AIR + 1)) * (((γ_AIR + 1)^2 * M^2)/((4 * γ_AIR * M^2) - 2*(γ_AIR - 1)))^(γ_AIR / (γ_AIR - 1))
    end
    return pt
end
